; BIOS-based Loader for the 32-bit MOOSE API.
; Copyright (c) Franois-Ren "Far" Bn Rideau Dang-Vu 1994

; Please Read the info/BIOS-API.nfo file

Ideal	; I can't bear Microsoftish things.
%NOCTLS
%NOLIST
%NOINCL
%MACS	; to see errors inside macros ; NOMACS is better when no errors :-|

; The LOADER will Load the API and jump to it.

; * First test the CPU Type. Do not load if not a 386 or better.
; * The LOADER detects if the API is in a DOS system file ( la MS-DOS), in
;  reserved sectors (the simplest trick), in extra sectors (as an invisible
;  Virus OS 8), or with its sectors mapped ( la LILO).
; * In case of failure, return with an error message and a "hit any key"
; * Get extended memory size from The BIOS.
; * Load the API Header in main memory.
; * Verify that there's enough main memory for it.
; * Load the API body in main memory.
; * Load the Kernel Header
; * Load the Kernel into extended memory (320 KB extended memory required)
; * Initialize the API fields; jump into the API.

; To Do:
; * Beware the current 28 KB limit for the API
; * Add the DOS loader ?
; * Have one different loader for each mode (H,C,D,M)
; * Add compression/decompression of the API & Kernel by the loader.
;  LHA has its self-decompressor fit in 1.5 KB, so we can do it too.

P386

;DEBUG=1
__NO_USE_INT_13H__=0

Include "../../macro.inc"
Include "../../mem386.inc"
Include "../../../M_Magic.h"	; magic numbers

; This really could have been written in C, if C could have produced small
; code that would have used only the BIOS.
%LIST

;;; Defines that may change.

; We mustn't really use these. The Boot Record need it, though.
LOADER_SEG_SIZE	 =  0F00h	;(60KB) ; size in para for the loader segment.
LOADER_STACK_TOP =  LOADER_SEG_SIZE*16	; SP in the same segment
LOADER_SECTS	 =  3			; size of the loader in sectors.
LOADER_OFFSET	 =  100h		; load the loader as a .COM file.
LOADER_UDATA	 =  0			; no udata area, as 80h-100h is used.
KERNEL_SECTS	 =  512			; Size of Kernel+API to load.
MIN_KERNEL_SPACE =  300			; KB needed by the kernel
MAX_KERNEL_SPACE =  16000		; KB asked by the kernel
MAX_TRY		 =  10		        ; number of retries before aborting I/O

;;; Defines that won't change
LOG_SECT_SIZE	=	9		; Size of sectors (low-level standard)
SECT_SIZE	=   1 shl LOG_SECT_SIZE	; 512
MAP_FILE	 =  LOADER_OFFSET+SECT_SIZE*LOADER_SECTS+LOADER_UDATA
			; just after the loader

; LOAD MODEs:
; * 'D' for loading a DOS file as the API+Kernel. LoadArg points to the
;  name (11 bytes) of the system file containing the API and the Kernel
; * 'H' for using an overformatted floppy's hidden sectors as the API+Kernel.
;  LoadArg contains real max tracks (hi) and real max sects/tracks (lo).
; * 'C' for using contiguous directories after the current one. LoadArg is
;  The number of the first sector (relative to hidden_sectors).
; * 'M' for LILO-like mapping. The map file is already loaded at address
;  MAP_FILE

model tiny
;; BUG in TASM 2.51: UDATA doesn't work :-(
;; masm;.data?;ideal doesn't work either: puts any data with a 32 bit address.
;; AAARRRGGGHHH
;; I'll use dataseg
Macro udata
  dataseg
  ;; If I could safely hook dataseg or other data segment declarations,
  ;; I'd save the old pointer here, and restore the udata pointer.
EndM udata

codeseg
		org	LOADER_OFFSET	; .COM file format
Entry_Point:
		jmp	near Init

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Loader DATA ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
		org	LOADER_OFFSET
;udata here (do not overwrite the jmp, thus '?' only)
CPUType 	dw	?		; AH:CPU, AL:NPU (see mem386.inc)
ErrorCode	db	?		; must be all on the same disk.

;udata	;; not only udata doesn't work, but I allocate data myself in the
	;;command line, thus reducing the disk size & memory size
ConvMem		equ	dword 80h	; Total Conventional memory in bytes
AvConvMem	equ	dword 84h	; Available Conv. mem in bytes
XtdMem		equ	dword 88h	; Total Extended memory in bytes
AvXtdMem	equ	dword 8Ch	; Available Extended mem in bytes

dataseg

BaseConvMem	dd	1000h  ;4096	; Page aligned, reserve 2K for BIOS.
BaseXtdMem	dd	110000h	;1MB+64KB	; Page aligned, reserve low
					; memory & HMA for dos emulator.

;;; Inits data for Next_cluster
Init_Sector_Read  equ	word 90h

;;; Returns the next block of sector to read in int 13h -able registers.
;;; al=0 if no more clusters.
Next_cluster	equ	word 92h


;;; we assume we could fill the user-defined buffer
;;; if not, then bufferize those routines with an overfillable
;;; buffer (overfilling: up to the maximal value given by Next_cluster).
over_fill	equ	dword 94h
over_fill_ofs	equ	word 94h
over_fill_seg	equ	word 96h

over_fill_len	dw	0

;;; We also assume no more than 64 KB overfilled buffer, no 64 KB boundary
;;; crossed, no pointer wraps, new value of buffer not interfering with old
;;; overfull buffer. (that's all :)

;;; Load Mode 'M'
pCluster	equ	word 98h

;;; Load Mode 'H' ; for fdformatted floppies only
;;; (can anyone achieve over-formatted hd's ?)
CurrentTrack	equ	word 9Ah	; hi: track, lo: head
Max_Tracks_on	equ	byte 9Ch	; max. number of "official" tracks

;;; Load Mode 'C'
CurrentSector	equ	dword 0A0h
MaxSector	equ	dword 0A4h

;;; Debugging LOADER
Int13Handler	equ	dword 0ACh

;;; Load Mode 'D'
DataArea	equ	word 0B0h
CurrentCluster	equ	word 0B2h
SectorInCluster	equ	word 0B4h
FAT_Reader	equ	word 0B6h	; routine to read 12 or 16 bit FAT
FAT_Mapped_Beg	equ	word 0B8h	; for 16 bit FATs, which part is mapped
FAT_Mapped_Len	equ	word 0BAh

;;; API Entry point
API_Entry_Point	equ	word		; in the API (0) segment

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ARGUMENTS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; see bootrec.asm for full details...

;;; DOS info table

bytes_per_sect	equ	word 0Bh
sectors_per_clu	equ	byte 0Dh
reserved_sect	equ	word 0Eh
num_FATs	equ	byte 10h
entries_rootdir	equ	word 11h
total_num_sect	equ	word 13h
media_descr	equ	byte 15h
sect_per_FAT	equ	word 16h
sect_per_track	equ	word 18h
num_heads	equ	word 1Ah
hidden_sect	equ	dword 1Ch
d_total_n_sect	equ	dword 20h

;;; MOOSE Loading Parameters

LoadMode	equ	byte 24h
LoadArg		equ	word 25h
BIOS_DL		equ	byte 27h
Clusters	equ	dword 28h

;;; API Header

API_Magic	equ	dword es:0h	; beware the override !
API_Length	equ	dword es:20h

;;; Kernel Header

Kernel_Magic	equ	dword es:0h
Kernel_Length	equ	dword es:20h


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Console I/O ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
codeseg

Putc	EQU	cPutc

; Adapted from MicroSoft's own boot sector
; DS:SI=ASCIIZ string to print
; AX spoilt, SI points just after the string-ending zero, 7 put into BX.
Proc	Puts	near
	cld
	lodsb
	or	al,al
	jz	Return
	PUTC
	jmps	Puts
EndP	Puts

IF 0
Proc	Beep
; Echo a BEEP to the the Console
	mov	al,7
;	jmps	cPutc
EndP	Beep
ENDIF

; Outputs a character to console
; spoils AX and DI, keeps si
Proc	cPutc
	push	bx	; no pusha: may be used on 8088 !!!
	push	si
	push	di
	push	es
	mov	ah,0Eh	; function Teletype
	mov	bx,7	; attribute
	int	10h	; Video BIOS call
	pop	es
	pop	di
	pop	si
	pop	bx
Return:
	retn
EndP	cPutc

@__

dataseg
  RegsMsg	db 13,10
		db "DI   SI   BP   SP   "
		db "BX   DX   CX   AX   "
		db "FLGS CS   DS   ES   "
		db "SS   IP   STK1 STK2"
		db 13,10,0
  ERegsMsg1	db 13,10
		db "EDI      ESI      EBP      ESP      "
		db "EBX      EDX      ECX      EAX      EFLAGS",13,10,0
  ERegsMsg2	db "CS   DS   ES   SS   IP   STK1 STK2 STK3 "
		db "STK4 STK5 STK6 STK7 STK8 STK9 STKA STKB"
		db 13,10,0
codeseg

Proc	DumpERegs
	push ss
	push es
	push ds
	push cs
	pushfd
	pushad
	ldea si,ERegsMsg1
	PUTS
	push ss
	pop es
	push cs
	pop ds
	mov di,sp
	mov cx,9
	call PutLZone
	ldea si,ERegsMsg2
	PUTS
	mov cx,16
	call PutWZone
	popad
	popfd
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
 Return2:
	ret
EndP	DumpERegs

Proc	DumpRegs
	push ss
	push es
	push ds
	push cs
	pushf
	pusha
	ldea si,RegsMsg
	PUTS
	mov di,sp
	push ss
	pop es
	push cs
	pop ds
	mov cx,16
	call PutWZone
	call PrCRLF
	popa
	popf
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
	ret
EndP	DumpRegs

Proc	PutLZone_
	call	Space
PutLZone:
	mov eax,[es:di]
	add di,4
	call PutHexDword
	loop PutLZone_
	ret
EndP	PutLZone_
Proc	PutWZone_
	call	Space
PutWZone:
	mov ax,[es:di]
	inc di
	inc di
	call	PutHexWord
	loop PutWZone_
	ret
EndP	PutWZone_
Proc	PutZone_
	call	Space
PutZone:
	mov al,[es:di]
	inc di
	call	PutHexByte
	jcxz	short Return2
	call	Space
	loop PutZone_
	ret
EndP	PutZone_

Proc	PutAddr
;; Prints the contents of es:di
	push	di
	mov	ax,es
	call	PutHexWord
	mov	al,':'
	call	Putc
	pop	di
	mov	ax,di
	jmp	PutHexWord
EndP	PutAddr

Proc	PutHexDword
	rol	eax,16
	call	PutHexWord
	rol	eax,16
;	jmp	PutHexWord
EndP	PutHexDword

PutHexWord:
; Prints the contents of AX in hexa.
; everything but F saved
		xchg	al,ah
		call	PutHexByte
		xchg	al,ah
;		jmp	PutHexByte

__@

; Prints the contents of AL in hexa.
; saves AX.
Proc	PutHexByte
	push	cx
	mov	cl,4
	rol	al,cl
	call	PutHexNybble
	rol	al,cl
	pop	cx
;	jmp	PutHexNybble
EndP	PutHexByte

; Prints to BIOS console the hexa nybble in low AL
; saves AX, puts 7 into BX
Proc	PutHexNybble
	push	ax
	and	al,0Fh
	add	al,'0'
	cmp	al,0Ah+'0'
	jb	do_putnyb
	add	al,'A'-'0'-0Ah
 do_putnyb:	
	PUTC
	pop	ax
	retn
EndP	PutHexNybble

Proc	cGetc
	clr	ah
	int	16h
	retn
EndP	cGetc

IF 0
Proc	sPutc
	stosb
	retn
EndP	sPutc
Proc	sGetc
	lodsb
	retn
EnP	sGetc
ENDIF

Proc	PrCRLF	near
	PUTC	13
	mov	al,10
	jmp	Putc
EndP	PrCRLF

@__
Proc	Space
	mov al,' '
	jmp Putc
EndP	Space
__@

;;;;;;;;;;;;;;;;;;;;;;;;;; Misc. Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
codeseg

Proc	safe_movsb
; compare source & destination
	@wo@	<13,10,"Moving safely ">,cx,<>
	@wo@	<" bytes from ">,ds,<":">
	@wo@	<>,si,<>
	@wo@	<" to ">,es,<":">
	@wo@	<>,di,<CRLF>
	push esi
	push edi
	push eax
	mov ax,ds
	movzx eax,ax
	shl eax,4
	movzx esi,si
	add esi,eax
	mov ax,es
	movzx eax,ax
	movzx edi,di
	shl eax,4
	add edi,eax
	cmp edi,esi
	pop eax
	pop edi
	pop esi
	ja sm_dest_gt_source
;sm_source_gt_dest:
	cld
	push	cx
	@IF ncxz
	  shr	cx,2
	@THEN
	  rep	movsd
	@ENDIF
	pop	cx
	@IF ncxz
	and	cx,3
	@THEN
	  rep	movsb
	@ENDIF
	ret
; assuming dest > source
 sm_dest_gt_source:
	add	si,cx
	add	di,cx
	sub	si,4
	sub	di,4
	std
	push	cx
	@IF ncxz
	  shr	cx,2
	@THEN
	  rep	movsd
	@ENDIF
	pop	cx
	add	si,3
	add	di,3
	@IF ncxz
	  and	cx,3
	@THEN
	  rep	movsb
	@ENDIF
	ret
EndP	safe_movsb

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; DISK PROCESSING ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

IF __NO_USE_INT_13H__ 
  Macro Int13h
    __ <"Int 13h called:">
    @_@
    pushf
    push cs
    call Do_Int13h
    __ <"Int 13h returned:">
    @_@
  EndM
  Proc Do_Int13h
    cli
    jmp [Int13Handler]
  EndP
ELSE
  Macro Int13h
    __ <"Int 13h called:">
    @_@
    Int 13h
    __ <"Int 13h returned:">
    @_@
  EndM
ENDIF


;;;;;;;;;;;;;;;;;;;;;;;;;;;; General Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc	SectorNumToBIOS_Args
; In: eax (0 for first sector)
; Out: BIOS sector identifier (cl,dh,ch)

	;__ <"SectorNumToBIOS_Args Input:">
	;@__@

	clr edx
	movzx ebx,[byte sect_per_track]
	div ebx
	mov cl,dl	; remaining: sector
	inc cl		; they are from one on
	movzx ebx,[byte num_heads]
	clr edx
	div ebx
	mov dh,dl	; remaining: head
	mov ch,al
	shl ah,6
	or cl,ah

	;__ <"SectorNumToBIOS_Args Output:">
	;@__@
	ret
EndP	SectorNumToBIOS_Args

;Proc	BIOSToSectorNum
; In: BIOS sector identifier
; Out: eax
;	dec cl			; sectors begin with 1
;	movzx ebx,cl		; ebx := sector number
;	and bl,3Fh
;	shr cl,6		; ax := cylinder number
;	mov al,ch
;	mov ah,cl
;	movzx ecx,dh		; ecx := head number
;	mul [byte num_heads]	; dx,ax := cylinder number * heads
;	push dx			; eax := dx,ax
;	push ax
;	pop eax
;	add eax,ecx
;	movzx edx,[byte sect_per_track]
;	mul edx
;	add eax,ebx
;	ret
;EndP	BIOSToSectorNum

Proc	Read_Sectors_With_Retries
;; ax,bx,cd,dx registers unchanged.
;; si (?), di spoilt
;; CF set if error (then see [ErrorCode])
 __ <"Reading Sectors">
 @_@
   mov di,MAX_TRY	; maximum number of retries
   mov ah,02h	; Read
  Read_Try:
   push di
   push ax
   Int13h
   mov [ErrorCode],ah
   pop ax
   jnc Read_ok
   push ax
   clr ah		; reset controller
   Int13h
   pop ax
   pop di
   dec di
   jnz Read_Try
 ;Too_Many_Retries:
   stc
   @by@ <"Disk I/O error 0x">,[ErrorCode],<>
   ret
 Read_ok:
   pop di
   ret
EndP	Read_Sectors_With_Retries

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc	Read_Once
; Input:
; * (Min) Length to read in ax
; * buffer in es:bx
; Output:
; * buffer (overwritten)
; * if buffer changed place, data moved, assuming no interference
; * ax: preserved
; * es:bx unchanged
; * cx: data actually read (overfull if > ax; end of file if < ax)
; *** NOTE: if you use data over bx+ax, you must set [over_fill_len] to
; ***  zero, or set it to another good value, and modify [over_fill_ofs]
; * other registers spoilt

 ;;; That's Structured programming in assembly !
 __ <"Read Once Input: ">
 @wo@ <"asked 0x">,ax,<" bytes at ">
 @wo@ <>,es,<":">
 @wo@ <>,bx,<CRLF>

 ;;; load old buffer address
   mov si,[over_fill_ofs]
   mov dx,[over_fill_seg]
   mov cx,[over_fill_len]
 @wo@ <"Already were 0x">,cx,<" bytes at ">
 @wo@ <>,dx,<":">
 @wo@ <>,si,<CRLF>

   jcxz RO_Buffer_ok
 ;;; compare to new one
   @IF nz	; If new buffer different
     mov di,es
     cmp di,dx
    ;@OR nz ; not supported yet
     jnz RO_new_buf_dif
     cmp si,bx
   @THEN	; Then move old to new
    RO_new_buf_dif:
     push ds
     push cx
     mov ds,dx
     mov di,bx
     call safe_movsb ; es:bx unchanged
     pop cx
     pop ds
   @ENDIF

 RO_Buffer_ok:
 __ <"Buffer now ok.">
   push bx
   @WHILE b	; While buffer strictly not overfull enough
     cmp cx,ax
   @DO		; Do fill it again
     push ax
     push cx
     push bx		; bx: buffer address
     call [Next_cluster]; al,cx,dx: next cluster to read.
     pop bx
     or al,al		; al: number of sectors to read
     jz End_Of_File
     push ax
     call Read_Sectors_With_Retries
     pop dx		; now, it's dl
     pop cx
     pop ax
     jc Disk_IO_Error
     __ <"Cluster Read">
     @_@
     shl dx,LOG_SECT_SIZE ; (!) wrong if LOG_ < 8
     add cx,dx
     add bx,dx
   @WEND
 ;; full enough, then give the caller what he asked
   pop bx ; original buffer
   pop ax ; original length asked for
   mov dx,cx ; length actually read.
   sub dx,ax
   mov [over_fill_len],dx
   mov dx,bx
   add dx,ax
   mov [over_fill_ofs],dx
   __ <"Done Reading once.">
   @_@
   ret

 End_Of_File:
   __ <"End of file reached">
   @_@
   pop cx ; total unsufficient length of data read
   pop ax ; length asked
   pop bx ; old buffer value
   mov dx,bx
   add dx,cx
   mov [over_fill_ofs],dx
   mov [over_fill_len],0
   ret

;;; On Errors,
 Disk_IO_Error:
   @PrByte <13,10,"Error 0x">,[ErrorCode]
   ldea si,db_zero
   pop bx ; original bx
   pop ax ; return address to Main
   jmp Exit
EndP	Read_Once

Proc DoReadOnce
   call Read_Once
   cmp cx,ax
   jb Premature_end_of_file
   ret
 Premature_end_of_file:
   lsa si,<13,10,"Unexpected end of file">
   pop ax; return address
   jmp near Exit
EndP DoReadOnce

;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'M' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc M_Init_disk near
    mov [pCluster],MAP_FILE
    ret
EndP M_Init_disk

Proc M_Next_cluster near
	push si
	mov si,[pCluster]
	lodsw
	mov cx,ax
	lodsw
	mov dh,ah
	@IF nz
	  or al,al
	@THEN
	  mov [pCluster],si
	@ENDIF
	pop si
	mov dl,[BIOS_DL]
	ret
;Entry format:
;BIOS_CX	dw	0002h		; CH:cyl&0xFF, CL:sect+((cyl>>8)<<6)
;BIOS_AL	db	02h		; Int 13h parameters to load the LOADER
;BIOS_DH	db	00h		; head
EndP M_Next_cluster

;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'D' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc D_Init_disk near
; Unsupported yet

; Look in the root directory, search for the system file, see length and
; cluster.

; ...

EndP D_Init_disk

Proc D_Next_cluster near

; memorized part of cluster number is next sectors to load.
; while we have a contiguous (without even track limit) zone of sectors on
; disk, look in the FAT for next cluster
; When we reach a limit (or when finished), load the contiguous part.
; memorize next part of cluster to load

; Look in the FAT for the cluster number
; if 12 bit FAT, all the FAT fits in buffer
; if 16 bit FAT, have bits of it; happily a single entry won't be split
; on two sectors.

; ...
EndP D_Next_cluster

;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'H' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The LoadParam is used as sect/trk (low) & #cyl (hi).

Proc H_Init_disk near
; Leave track 0 head 0 for the boot sector + fake boot sector + loader
; floppies have less than 255 tracks and 2 heads, (thus) less than 65535 sects.
   mov [word CurrentTrack],0100h ; already read Cyl 0 (lo), Hd 1 (hi)
   mov ax,[total_num_sect]
   shr ax,1 ; number of heads=1 ; cf should be 0
   div [byte sect_per_track]
   ; we should have ah=0 (all sectors used in declared surfaces&cylinders)
   mov [Max_Tracks_on],al
   ret
EndP H_Init_disk

Proc H_Next_cluster near
   mov ax,[LoadArg]	; real max tracks & max sects/tracks
   mov dx,[CurrentTrack]
   @IF z
     xor dh,1	; change head
   @THEN
     inc dl	; change track
   @ENDIF
   cmp dl,[Max_Tracks_on]
   ja  extra_track
   mov cl,[byte sect_per_track]
   sub al,cl
   inc cl
 H_done:
   mov [CurrentTrack],dx
   mov ch,dl
   mov dl,[BIOS_DL]
   ret
 extra_track:
   mov cl,1
   cmp dl,ah	; only ah real tracks.
   jbe H_done
   clr al
   ret
EndP H_Next_cluster


;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'C' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc C_Init_disk near
	__ <"'C' Init disk">
	mov ebx,[hidden_sect]
  ; we store the next sector to read in CurrentSector
	movzx eax,[LoadArg]
	add eax,ebx
	mov [CurrentSector],eax
  ; we store the maximal value to read in MaxSector
	movzx eax,[reserved_sect]
	add eax,ebx
	mov [MaxSector],eax
	;@__@
	ret
EndP C_Init_disk

Proc C_Next_cluster near
	clr al
	mov esi,[CurrentSector]
	mov edi,[MaxSector]
	sub edi,esi
	jb CNcret
	push esi
	mov eax,esi
	call SectorNumToBIOS_Args
	mov dl,cl ; sector
	and dl,3Fh
	mov al,[byte sect_per_track]
	sub al,dl
	inc al
	movzx eax,al
	@IF a
	  cmp eax,edi
	@THEN
	  mov eax,edi
	@ENDIF
	pop esi
	add esi,eax
	mov [CurrentSector],eax
	mov dl,[BIOS_DL]

;__ <"'C' Next Cluster Output:">
	;@__@

  CNcret:
	ret
EndP C_Next_cluster


;;;;;;;;;;;;;;;;;;;;;;;;;; Low-Level Disk Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;

; From Linux:
; * This procedure turns off the floppy drive motor, so
; * that we enter the kernel in a known state, and
; * don't have to worry about it later.
Proc kill_motor near
	push dx
	mov dx,03f2h
	xor al,al
	out dx,al
	pop dx
	retn
EndP kill_motor

;;;;;;;;;;;;;;;;;;;;;;;;; Initialization Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;; These procedures are executed only once, so I macro'ed them.
Macro	Init_LOADER
;;; Test CPU & NPU type
  def_proc_TestCPU inline
  mov [CPUType],ax
  @IF b
    cmp ah,3
  @THEN
    lsa si,<'MOOSE Needs a 32 bit processor'>
    jmp near Exit
  @ENDIF
  @__
    push es
    push 0		; Arg ! How to tell to use long or word ???
    pop es
    mov eax,[es:13h*4]
    mov [Int13Handler],eax
    pop es
  __@
EndM	Init_LOADER

Macro	Parse_Options
;;; Verify that DOS aint loaded
   @IF z
	cmp [byte 80h],99h	; The bootrec ought to have put this value
				; 99h >=80h, thus COMMAND.COM won't put it.
   @THEN
	mov al,[LoadMode]
	;ldea bx,D_Init_disk
	;ldea dx,D_Next_cluster
	cmp al,'D'
	jz Parsed
	ldea bx,M_Init_disk
	ldea dx,M_Next_cluster
	cmp al,'M'
	jz Parsed
	ldea bx,C_Init_disk
	ldea dx,C_Next_cluster
	cmp al,'C'
	jz Parsed
	ldea bx,H_Init_disk
	ldea dx,H_Next_cluster
	cmp al,'H'
	jz Parsed

	dataseg
	  Mode_not_supported	db "Load Mode "
	  Unsupported_Mode	db 'X'," not supported."
	  db_zero		db 0
	codeseg
	mov [Unsupported_Mode],al
	ldea si,Mode_not_supported
	jmp near Exit
   @ELSE
      lsa si,<"I won't work from DOS; Load me directly with bootrec.bin.">
      jmp near Exit
   @ENDIF
   Parsed:
	mov [Init_Sector_Read],bx
	mov [Next_cluster],dx
EndM	Parse_Options

Macro	Init_Message	near
	PUTS	'Loading..'
EndM	Init_Message

Macro	Get_Memory
  ;;; Get Total extended memory
	mov ah,88h
	int 15h		; result in KB
	movzx eax,ax
	shl eax,10
	mov [XtdMem],eax; result in bytes
  ;;; Compute available x. mem.
	sub eax,65536	; Don't put anything in the HMA zone
	mov [AvXtdMem],eax
  ;;; Get Total conventional memory
	int 12h		; result in KB
	movzx eax,ax
	shl eax,10
	mov [ConvMem],eax
  ;;; Compute Available conventional memory
    ;- assuming the boot record loaded us as high as possible)
    ;- assuming the stack has only one word element (the .COM ending return
    ; address), and is the topmost free memory. (higher are boot-installed
    ; BIOS residents).
    ;- reserving the first page of memory for BIOS/VM86 DOS use.
	mov ax,ss
	movzx eax,ax
	movzx esp,sp
	shl eax,4
	lea eax,[eax+esp+2-4096]
	mov [AvConvMem],eax
EndM	Get_Memory

Macro	Init_Disk_Read
   ;;; Init sector count
	call [Init_Sector_Read]
EndM	Init_Disk_Read


Macro	LOAD_API
   ; Note: for the moment, the API should not be larger than 60 KB
   ;  Than even 28 KB to avoid a 64KB DMA limit with sector prefetching !
   ;  If 28 KB is to be overflown, we must change the memory access code.

   ;;; Init The buffer
	mov ax,0100h
	mov es,ax	; 4KB
	clr bx
   ;;; Load the Header
	mov ax,64	; Header length
	call DoReadOnce
   ;;; Get the length from the header
	@IF nz
	  cmp [API_Magic],MAGIC_PC_API
	@THEN
	  lsa si,<"API file fails magic number test.">
	  jmps Exit
	@ENDIF
	mov eax,[API_Length]
	@IF a
	  cmp eax,[AvConvMem]
	@THEN
	  lsa si,<"Not enough memory for API">
	  jmps Exit
	@ENDIF
	sub eax,64
   ;;; Load the entire API.
   ;;; ... ToDo:
   ;;; - add relocation (?)/ header parsing
   ;;; - add decompression (?)
; ...........
;   ReadOnce
EndM	LOAD_API

Macro	LOAD_MOOSE
 ; ....
EndM	LOAD_MOOSE

Macro	Set_MOOSE_Options
 ; ....
EndM	Set_MOOSE_Options

Macro	Go_MOOSE
 ; ....
 PUTS <13,10,"Going MOOSE !">
EndM	Go_MOOSE

Macro	Do_Exit
	call PrCRLF
	PUTS
	PUTS	<13,10,'Hit any key to reboot'>
	GETC
EndM	Do_Exit

Proc	Init	near
 ;__ <"Testing @_@">
	;@_@

 __ <"Initing self: CPU Test:">
	Init_LOADER
	@wo@ <>,[CPUType],<>

 __ <"Testing Int 13h">
	push 0
	pop es
	@addr@ <" at ">,<[word es:13h*4+2]>,<[word es:13h*4]>,<>
	mov ah,0
	mov dl,0
	Int13h
	@_@

 __ <"Looking at options.">
	Parse_Options

 __ <"Hello, World !">
	Init_Message

 __ <"Seeing how much memory is left.">
	Get_Memory
	@dwo@ <13,10,"Found 0x">,[AvConvMem],<" bytes of conventional memory,"
	@dwo@ <13,10,"and 0x">,[AvXtdMem],<" bytes of extended memory.">

 __ <"Initialize disk reading">
	Init_Disk_Read

 __ <"Load the API at 4KB">
	LOAD_API
	@dwo@ <"Done loading a 0x">,[API_Length],<" long API.">

 __ <"Load MOOSE at 1MB+64K">
	LOAD_MOOSE

 __ <"Set options for MOOSE">
	Set_MOOSE_Options

 __ <"Going MOOSE">
	Go_MOOSE		; Let's GO !

Exit:
 __ <"Aborting...">
	Do_Exit			; On error, exit
	ret
EndP	Init

;;;;;; Initialization only data ;;;;;;

dataseg
  align 4
  BufferBeg:

End Entry_Point
