#ifndef _LLOBJECTS_H
#define _LLOBJECTS_H

/*
/ llobjects.h
/   contains the definition of low-level classes and objects under
/ the pcore implementation of OSL.
/
*/

/************************* TYPDEF's *******************************/

typedef union pcore_cell		pcore_cell ;
typedef struct pcore_ll_object		pcore_ll_object ;
typedef struct pcore_ll_descriptor	pcore_ll_descriptor ;
typedef struct pcore_name		pcore_name ;
typedef struct pcore_identifier		pcore_identifier ;
typedef struct pcore_continuation	pcore_continuation ;
typedef struct pcore_attribute		pcore_attribute ;
typedef struct pcore_paramed_fun	pcore_paramed_fun ;
typedef pcore_cell	pcore_fun		(pcore_cell,...) ;
typedef pcore_cell	pcore_fun2		(pcore_cell,pcore_cell,...) ;


/*************** Low Level objects are made up of cells. ***********/
union pcore_cell {
 /* terminal data view point */
  unsigned long		_ulong ;
  signed long		_slong ;
  unsigned int		_uint ;
  signed int		_sint ;
  signed char		_schar ;
  unsigned char		_uchar ;
  size_t		_size ;
/*  float		_float ; */

 /* pointer view point */
  void *		_voidp ;
  unsigned char *	_ucharp ;
  pcore_cell *		_cellp ;
  pcore_ll_object *	_llobjp ;
  pcore_ll_descriptor * _lldescp ;
  pcore_fun *		_funp ;
  pcore_fun2 *		_fun2p ;
};


/************************* Parametered function **************************/
struct pcore_paramed_fun {
  pcore_fun *	fp ;
  pcore_cell	param ;
};


/********************** object description structures *********************/

/**** Low level ****/
struct pcore_ll_descriptor {
  pcore_fun *	size ;		/* get the size of the object */
  pcore_fun *	copy ;		/* copy the object's value */
  pcore_fun *	link ;		/* create a hard link on the object */
  pcore_fun *	unlink ;	/* destroy a hard link on the object */
  pcore_fun *	compare ;	/* compare the object to another */
};


/**** Medium level ****/
struct pcore_ml_descriptor {
  pcore_fun *	recurse ;	/* go recursively into the object (1) */
  pcore_fun *	deepcopy ;	/* see (2) */
  pcore_fun *	deepcompare ;	/* see (3) */
};

struct ll_object {
  pcore_ll_descriptor *		descriptor ;
  pcore_cell			data ;
};

struct pcore_identifier {
  pcore_name *	name ;

};

struct pcore_attribute {
  pcore_identifier *	name ;
  pcore_cell		(*value)	(pcore_ll_object) ;
 /* ... */
};

struct pcore_name {
  char * shortname ;		/* just the name */
};


#endif

/*
Notes:
=====

* Recursing through an object's structure:
apply the function recurse of the object's descriptor, with parameters:
- the object
- a to-apply parametered function, that will be called with sub-element
parameters.
* A to-apply function will have to detect if the node is terminal; then
 it'll call the element with the recurse function again, or with the
 first parameter's described paramed-fun
***/
